/***************************************************************************
 multiple.cc  -  description
 -------------------
 begin                : Fri Dec 10 1999
 copyright            : (C) 1999 by Juan Pablo Romero Mndez
 email                : juan_pablo_22@yahoo.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/



#include "multiple.h"
#include "object.h"
#include "read_data.h"

#include "util.h"
#include "trackball.h"

#include <stdlib.h>
#include <strstream>
#include <algo.h>
#include <pair.h>

// static members
// 

gint       Multiple::_selection_name_stack = 1;
gboolean   Multiple::_select_mode          = FALSE;
GnodeList  Multiple::_selection_gnode_list;

deque<GnodeList> prev_parents_list;
deque<GnodeList>::iterator parIter;

// 

Multiple::Multiple()
{
    _type = MULTIPLE;
}

static void
_delete_node(GNode* gnode, gpointer data)
{
    delete BASE_OBJECT(gnode);
    g_node_destroy(gnode);
}

Multiple::~Multiple()
{
    g_node_children_foreach( _gnode, G_TRAVERSE_ALL, _delete_node, NULL );
    g_node_destroy(_gnode);
}


// insert a new object into the tree
// this function is called _only_ with newly create objects.
// in other cases, use move()

GNode*
Multiple::append(BaseObject* baseob)
{
    g_assert( baseob != this );
    baseob->_gnode = g_node_append_data(_gnode, baseob);

    // the last element of the deque must point to baseob->_gnode
    // because that transformation correspond to the node itself

    if( Object* ob = dynamic_cast <Object*>( baseob ) )
        ob->_trans.back()._node = baseob->_gnode;

    return baseob->_gnode;
}



void
Multiple::remove(GNode* node)
{
    if( !node )
        return;

    BaseObject * parent =  BASE_OBJECT(node->parent);

    // destroy the object and the children (if any)
    delete node->data;

    // destroy the gnode
    g_node_destroy(node);

    // the parent must recalculate the bounding box
    parent->update_bbox();
}


static gboolean
_create_prev_parents(GNode* node, gpointer data)
{
    if( dynamic_cast<Multiple*>( BASE_OBJECT(node) ) )
        return FALSE;

    prev_parents_list.push_back( GnodeList() );

    GNode* ptr = node->parent;

    // build the deque of the original parents
    // elem # 0 is the root of the tree.
    while( ptr ) {
        prev_parents_list.back().push_front( ptr );
        ptr = ptr->parent;
    }

    return FALSE;
}

static gboolean
_update_transf_list(GNode* node, gpointer data)
{
    if( dynamic_cast<Multiple*>( BASE_OBJECT(node) ) )
        return FALSE;

    Object* ob  = (Object*) node->data;
    GNode*  ptr = node->parent;
    GnodeList   new_parents;

    // build the deque of the new parents
    while( ptr ) {
        new_parents.push_front( ptr );
        ptr = ptr->parent;
    }

    // find the position where the two lists begin to mismatch
    pair<Iter, Iter> begin_mismatch = mismatch( parIter->begin(),
                                                parIter->end(),
                                                new_parents.begin() );

    // position where the two lists end to mismatch
    pair<rIter, rIter> rev_mismatch = mismatch( parIter->rbegin(),
                                                parIter->rend(),
                                                new_parents.rbegin() );

    GnodeList new_nodes;
    insert_iterator<GnodeList> ii(new_nodes, new_nodes.begin());

    // these are the nodes that will be inserted in the list
    copy( begin_mismatch.second, rev_mismatch.second.base(), ii);


    // get the transformation node that corresponds to the <GNode*> : begin_mismatch.first
    // It is the first element that do not match from the old parents list

    trIter ti, tj;

    for( ti = ob->trans().begin(); ti != ob->trans().end(); ti++ ) {
        if( ti->node() == *(begin_mismatch.first) )
            break;
    }

    // set '_node' to NULL to the transformations that doesn't have an asocciated
    // 'Multiple' node.

    for( tj = ti; tj->node() != *(rev_mismatch.first - 1); tj++ )
        tj->node( NULL );

    // insert the new nodes in the transf. list
    // ( _before_ 'begin_mismatch.first').

    for( Iter i = new_nodes.begin(); i != new_nodes.end(); i++) {
        ob->trans().insert(ti++, Transformations( *i ));
    }

    // select the next prev_parent_list;
    parIter++;

    return FALSE;
}


static GNode*
_move_1_node(GNode* node, GNode* root_node)
{
    BaseObject* parent = BASE_OBJECT(node->parent);
    Object*     ob     = (Object*) node->data;

    GNode* ptr = node->parent;
    GnodeList prev_parents, new_parents;

    // build the deque of the original parents
    // elem # 0 is the root of the tree.
    while( ptr ) {
        prev_parents.push_front( ptr );
        ptr = ptr->parent;
    }

    // unlink the node
    g_node_destroy(node);
    parent->update_bbox();

    // insert the object to the new location
    ob->gnode( g_node_append_data(root_node, ob) );

    // new parent
    ptr = ob->gnode()->parent;

    // build the deque of the new parents
    while( ptr ) {
        new_parents.push_front( ptr );
        ptr = ptr->parent;
    }

    pair<Iter, Iter> begin_mismatch = mismatch(prev_parents.begin(),
                                               prev_parents.end(),
                                               new_parents.begin());

    pair<rIter, rIter> rev_mismatch = mismatch(prev_parents.rbegin(),
                                               prev_parents.rend(),
                                               new_parents.rbegin());



    GnodeList new_nodes;
    insert_iterator<GnodeList> ii(new_nodes, new_nodes.begin());

    copy( begin_mismatch.second, rev_mismatch.second.base(), ii);


    //get the transformation node that corresponds to the <GNode*> : begin_mismatch.first
    // It is the first element that do not match
    trIter ti, tj;
    Iter it;

    for( ti = ob->trans().begin(); ti != ob->trans().end(); ti++ ) {
        // begin_mismatch.first isn't valid when prev_parents is a subset
        // of new_parents (as in A...X, A...XY), because it points to the
        // last element plus one of the first list.
        // begin_mismatch.first-1 is always valid
        if( ti->node() == *(begin_mismatch.first-1) ) {
            ti++;
            break;
        }
    }

    // set '_node' to NULL to the transformations that doesn't have an asocciated
    // 'Multiple' node.
    for( tj = ti, it = begin_mismatch.first;
         tj->node() != *(rev_mismatch.first - 1) &&  it != prev_parents.end();
         tj++, it++ )
        tj->node( NULL );

    // insert the new nodes in the transf. list
    // (before 'begin_mismatch.first').
    for( Iter i = new_nodes.begin(); i != new_nodes.end(); i++) {
        ob->trans().insert(ti++, Transformations( *i ));
    }

    return ob->gnode();
}

// TODO: Checar esta funcin!!
// Si todo funciona bin, '_move_1_node' se puede poner en trminos
// de '_update_transf_list'

GNode*
Multiple::move(GNode* node)
{
    BaseObject* parent = BASE_OBJECT(node->parent);
    BaseObject* ob     = BASE_OBJECT(node);
    Multiple* multiple = dynamic_cast <Multiple*>( ob );
    GNode* ret;

    if( multiple ) {

        // clear the list
        deque<GnodeList>().swap(prev_parents_list);
        g_node_traverse(_gnode,G_PRE_ORDER,G_TRAVERSE_LEAFS,-1,_create_prev_parents,NULL);

        // unlink the node
        g_node_destroy(node);
        parent->update_bbox();
        // insert the object to the new location (this node)
        ob->_gnode = g_node_append_data(_gnode, ob);

        // reset the iter position
        parIter = prev_parents_list.begin();
        g_node_traverse(_gnode,G_PRE_ORDER,G_TRAVERSE_LEAFS,-1,_update_transf_list,NULL);

        ret = ob->_gnode;
    }
    else {
        ret = _move_1_node(node, _gnode);
    }

    calc_center();
    update_bbox();
    return ret;
}

void
Multiple::unlink(GNode* node)
{
    BaseObject* parent = BASE_OBJECT(node->parent);
    g_node_destroy(node);
    parent->update_bbox();
}

void
Multiple::parse_xml_node(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur) {}

void
Multiple::initialize()
{
    static guint _loaded = 0;
    ostrstream n;
    n << "multiple" << _loaded++ << ends;
    _name = n.str();
    n.freeze(0);

}


BaseObject *
Multiple::part(gint n) {
    g_assert(n >= 0);
    return BASE_OBJECT( g_node_nth_child(_gnode, n) );
}


static gboolean
_find_selected(GNode* node, gboolean * found)
{
    if( BASE_OBJECT(node)->selected() ) {
        *found = TRUE;
        return TRUE;
    }

    return FALSE;
}

// True only if some child (or itself) is selected

gboolean
Multiple::some_selected()
{
    gboolean found = FALSE;
    g_node_traverse(_gnode, G_PRE_ORDER, G_TRAVERSE_ALL, -1, _find_selected, &found);

    return found;
}


static gboolean
_find_motion(GNode* node, gboolean * found)
{
    if( BASE_OBJECT(node)->motion() ) {
        *found = TRUE;
        return TRUE;
    }

    return FALSE;
}

// True only if some child is in motion

gboolean
Multiple::some_motion()
{
    gboolean found = FALSE;
    g_node_traverse(_gnode, G_PRE_ORDER, G_TRAVERSE_ALL, -1, _find_motion, &found);

    return found;
}


gboolean
_find_idle(GNode* node, gboolean* found)
{
    if( BASE_OBJECT(node)->idle() ) {
        *found  = TRUE;
        return TRUE;
    }

    return FALSE;
}

// True only if some child *or* itself has idle set

gboolean
Multiple::some_idle()
{
    gboolean found = FALSE;
    g_node_traverse(_gnode, G_PRE_ORDER, G_TRAVERSE_ALL, -1, _find_idle, &found);

    return found;
}

gboolean
_find_child(GNode* node, Object** child)
{
    if( Object* ob = dynamic_cast<Object*>(BASE_OBJECT(node)) ) {
        *child = ob;
        return TRUE;
    }

    return FALSE;
}



Object*
Multiple::some_child()
{
    Object* child = NULL;
    g_node_traverse(_gnode, G_PRE_ORDER, G_TRAVERSE_ALL, -1, _find_child, &child);
    return child;
}



static gboolean
_offset(GNode* node, pair<GNode*, gfloat*> * data)
{
    if( Object* ob = dynamic_cast<Object*>( BASE_OBJECT(node) ) ) {
        Transformations* tr = ob->get_transf_by_node(data->first);
        tr->offset()[0] += data->second[0];
        tr->offset()[1] -= data->second[1];
    }

    return FALSE;
}


void
Multiple::offset  (gfloat * off)
{
    g_node_traverse( _gnode,
                     G_PRE_ORDER,
                     G_TRAVERSE_LEAFS, -1,
                     _offset,
                     & make_pair(_gnode, off) );
}

gboolean
_lastquat(GNode* node, pair<GNode*, gfloat*> * data)
{
    if( Object* ob = dynamic_cast<Object*>( BASE_OBJECT(node) ) ) {
        Transformations* tr = ob->get_transf_by_node(data->first);
        tr->lastquat(data->second);
    }
    return FALSE;
}



void
Multiple::lastquat  (gfloat * quat)
{
    g_node_traverse( _gnode,
                     G_PRE_ORDER,
                     G_TRAVERSE_LEAFS,-1,
                     _lastquat,
                     & make_pair(_gnode, quat) );
}


static gboolean
_curquat(GNode* node, pair<GNode*, gfloat*> * data)
{
    if( Object* ob = dynamic_cast<Object*>( BASE_OBJECT(node) ) ) {
        Transformations* tr = ob->get_transf_by_node(data->first);
        tr->curquat(data->second);
    }
    return FALSE;
}

void
Multiple::curquat  (gfloat * quat)
{
    g_node_traverse( _gnode,
                     G_PRE_ORDER,
                     G_TRAVERSE_LEAFS,-1,
                     _curquat,
                     & make_pair(_gnode, quat) );
}

static gboolean
_center(GNode* node, pair<GNode*, gfloat*> * data)
{
    if( Object* ob = dynamic_cast<Object*>( BASE_OBJECT(node) ) ) {
        Transformations* tr = ob->get_transf_by_node(data->first);
        tr->center(data->second);
    }
    return FALSE;
}

void
Multiple::center  (gfloat * c)
{
    g_node_traverse( _gnode,
                     G_PRE_ORDER,
                     G_TRAVERSE_LEAFS,-1,
                     _center,
                     & make_pair(_gnode, c) );
}


static void
gnode_draw(GNode* gnode, gpointer _select)
{
    BaseObject * baseob = BASE_OBJECT(gnode);
    if( baseob->invisible() )
        return;

    Object* ob = dynamic_cast <Object*>( baseob );

    if ( ob != NULL ) {
        glPushMatrix();
        if ( GPOINTER_TO_INT(_select) ) {
            // only non-multiple objects gets a name
            ob->gl_load_name( Multiple::_selection_name_stack ); //TODO: Sirve para algo esto?
            Multiple::_selection_gnode_list.push_back(gnode);
            glLoadName(Multiple::_selection_name_stack++);
        }
        ob->apply_transformations();
        ob->draw();
        glPopMatrix();
    }
    else
        baseob->draw();

}

void
Multiple::draw(void)
{
    Object* ob;
    // Draw the bbox
    /*
    if( show_bbox() && !_select_mode && (ob = some_child()) ) {
        Transformations* tr = ob->get_transf_by_node(_gnode);
        glPushMatrix();
        glTranslatef( tr->_offset[0],
                      tr->_offset[1],
                      tr->_offset[2] );
        rotate_about( *tr );
        draw_bbox();
        glPopMatrix();
    }
    */
    g_node_children_foreach( _gnode,
                             G_TRAVERSE_ALL,
                             gnode_draw,
                             GINT_TO_POINTER(_select_mode) );
}

static gboolean
_find_nodes(GNode* gnode, deque<Object*> * lst)
{
    if( Object* ob = dynamic_cast<Object*>( BASE_OBJECT(gnode) ) )
        lst->push_back(ob);

    return FALSE;
}

static gboolean
_calc_bbox(GNode* gnode, gpointer data)
{
    deque<Object*> childs;
    deque<Object*>::iterator it;
    SVector<6> bbox;
    Multiple* multiple = (Multiple*) gnode->data;

    g_node_traverse(gnode, G_PRE_ORDER, G_TRAVERSE_LEAFS, -1, _find_nodes, & childs);

    if( childs.size() == 0 )
        return FALSE;

    gint d = g_node_depth(gnode) - 1;

    multiple->bbox( childs[0]->bboxes()[d] );

    for( it = childs.begin(); it != childs.end(); it++ ) {

        //bbox = it->bboxes()[d];
        bbox = (*it)->bboxes()[d];

        for(gint i=0; i < bbox.size(); i++) {
            if( i % 2 ) { // odd (min_*)
                if( bbox[i] < multiple->bbox()[i] )
                    multiple->bbox()[i] = bbox[i];
            }
            else { // even (max_*)
                if( bbox[i] > multiple->bbox()[i] )
                    multiple->bbox()[i] = bbox[i];
            }
        }
    }

    return FALSE;
}


static gboolean
_update_bboxes(GNode* gnode, gpointer data)
{
    GNode* ptr = gnode->parent;

    if( Object* ob = dynamic_cast<Object*>( BASE_OBJECT(gnode) ) )
        ob->update_bboxes();

    return FALSE;
}

// find the union of all the bounding boxes

void
Multiple::update_bbox()
{
    if( is_empty() )
        return;

    g_node_traverse(_gnode,G_PRE_ORDER,G_TRAVERSE_LEAFS,-1,_update_bboxes, NULL);

    g_node_traverse(_gnode,G_PRE_ORDER,G_TRAVERSE_NON_LEAFS,-1,_calc_bbox, NULL);
}




SVector<3> &
Multiple::center()
{
    static SVector<3> c;
    if( is_empty() )
        return c;

    GNode* ptr = _gnode->children;

    // find a leaf node
    while( ptr->children )
        ptr = ptr->children;

    
    if( Object* ob = dynamic_cast<Object*>( BASE_OBJECT(ptr) ) )
        return ob->get_transf_by_node(_gnode)->_center;
    else
        return c;
}


static void
_update_center(GNode * gnode, Multiple * multiple)
{
    BaseObject* ob = BASE_OBJECT(gnode);
    ob->calc_center();
}

static gboolean
_set_center(GNode * gnode, pair< GNode*, SVector<3> > * data)
{
    if( Object* ob = dynamic_cast<Object*>(BASE_OBJECT(gnode)) )
        ob->get_transf_by_node(data->first)->center( data->second );

    return FALSE;
}

// Calculate the vertex mean,
// applying the function to the nodes as needed.
void
Multiple::calc_center()
{
    if( is_empty() )
        return;

    g_node_children_foreach( _gnode,
                             G_TRAVERSE_ALL,
                             _update_center,
                             NULL );
    gint n = length();
    SVector<3> center;

    while( n-- > 0 )
        center += part(n)->center();

    // Average
    times(& center, & center, 1.0/length());

    g_node_traverse( _gnode,
                     G_PRE_ORDER,
                     G_TRAVERSE_LEAFS, -1,
                     _set_center,
                     & make_pair(_gnode, center) );
}

gfloat
Multiple::farest_vertex(gfloat* center)
{
    gfloat tmp;
    gint n = length();
    gfloat radio = 0;
    
    while( n-- > 0 ) {
        tmp = part(n)->farest_vertex(center);
        if( radio < tmp )
            radio = tmp;
    }
    
    return radio;
}


static void
_set_show_bbox_all(GNode* gnode, gpointer data) {
    BASE_OBJECT(gnode)->show_bbox(GPOINTER_TO_INT(data));
}

void
Multiple::show_bbox_all(gboolean s)  {
    _show_bbox = s;
    g_node_children_foreach( _gnode, G_TRAVERSE_ALL, _set_show_bbox_all, GINT_TO_POINTER(s));
}


static void
_set_idle_all(GNode* gnode, gpointer data) {
    BASE_OBJECT(gnode)->idle( GPOINTER_TO_INT(data) );
    Multiple *multiple = dynamic_cast <Multiple*>( BASE_OBJECT(gnode) );
    if( multiple )
        multiple->idle_all( GPOINTER_TO_INT(data) );
}

void
Multiple::idle_all(gboolean s)  {
    _idle = s;
    g_node_children_foreach( _gnode, G_TRAVERSE_ALL, _set_idle_all, GINT_TO_POINTER(s));
}



//  some virtual functions //
// This macros, declare the functions needed to apply an attribute to
// all the children of _gnode;
//
// The second is used only for gfloat attributes.

#define VIRTUAL_ATTRIBUTE_IMPL(name, type, setargs, foreachargs)\
    static void _set_##name(GNode* gnode, gpointer data) {      \
    BASE_OBJECT(gnode)->name(setargs);                          \
    }                                                           \
    void Multiple::name(type s)  {                              \
        _##name = s;                                            \
    g_node_children_foreach( _gnode, G_TRAVERSE_ALL, _set_##name, foreachargs); \
    }


#define VIRTUAL_ATTRIBUTE_IMPL_FLOAT(name) \
    static void _set_##name(GNode* gnode, gfloat* val) {     \
    BASE_OBJECT(gnode)->name(*val);                          \
    }                                                        \
    void Multiple::name(gfloat s)  {                         \
    gfloat val[1];                                           \
    val[0] = s;                                              \
        _##name = s;                                         \
    g_node_children_foreach( _gnode, G_TRAVERSE_ALL, _set_##name, val); \
    }


VIRTUAL_ATTRIBUTE_IMPL(smooth,       gboolean, GPOINTER_TO_INT(data), GINT_TO_POINTER(s));
VIRTUAL_ATTRIBUTE_IMPL(mono,         gboolean, GPOINTER_TO_INT(data), GINT_TO_POINTER(s));
VIRTUAL_ATTRIBUTE_IMPL(normals,      gboolean, GPOINTER_TO_INT(data), GINT_TO_POINTER(s));
VIRTUAL_ATTRIBUTE_IMPL(transparent,  gboolean, GPOINTER_TO_INT(data), GINT_TO_POINTER(s));
VIRTUAL_ATTRIBUTE_IMPL(show_mesh,   gboolean, GPOINTER_TO_INT(data), GINT_TO_POINTER(s));
VIRTUAL_ATTRIBUTE_IMPL(mesh_smooth, gboolean, GPOINTER_TO_INT(data), GINT_TO_POINTER(s));
VIRTUAL_ATTRIBUTE_IMPL(smooth_lines, gboolean, GPOINTER_TO_INT(data), GINT_TO_POINTER(s));
VIRTUAL_ATTRIBUTE_IMPL(smooth_edges, gboolean, GPOINTER_TO_INT(data), GINT_TO_POINTER(s));


VIRTUAL_ATTRIBUTE_IMPL(mono_color,        gdouble*,     (gdouble*) data, s);
VIRTUAL_ATTRIBUTE_IMPL(polygon_mode,      PolygonMode, (PolygonMode) GPOINTER_TO_INT(data), GINT_TO_POINTER(s));

VIRTUAL_ATTRIBUTE_IMPL_FLOAT(frame_line_width);
VIRTUAL_ATTRIBUTE_IMPL_FLOAT(lines_width);
VIRTUAL_ATTRIBUTE_IMPL_FLOAT(points_width);

