/***************************************************************************
                          object.cc  -  description
                             -------------------
    begin                : Fri Dec 10 1999
    copyright            : (C) 1999 by Juan Pablo Romero Mndez
    email                : juan_pablo_22@yahoo.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "object.h"

#include "read_data.h"
#include "util.h"
#include "trackball.h"
#include "sarray.h"

#include <math.h>
#include <string.h>
#include <GL/gl.h>


PolygonMode Object::_polygon_mode_down = POLY_NONE;


Object::Object()
{
    _trans.push_front( Transformations() );
}

Object::~Object() {}


Transformations*
Object::get_transf_by_node(GNode* n)
{
    deque<Transformations>::iterator ti;

    gint s = _trans.size();

    for( ti = _trans.begin(); ti != _trans.end(); ti++ ) {
        if( ti->_node == n )
            return &(*ti);
    }

    g_error("GNode: 0x%x isn't in the transform. list", n);
}

void
Object::delete_empty_transformations()
{
    deque<Transformations>::iterator it;

    for( it = _trans.begin(); it != _trans.end(); it++ )
        if( it->_lastquat == SVector<4>(0,0,0,1) &&
            it->_offset == SVector<3>()
          )
            _trans.erase(it);
}


void
Object::offset(gfloat * off)
{
    _trans.back()._offset[0] += off[0];
    _trans.back()._offset[1] -= off[1];
}

void
Object::lastquat(gfloat * quat)
{
    _trans.back().lastquat(quat);
}

void
Object::curquat(gfloat * quat)
{
    _trans.back().curquat(quat);
}

void
Object::center(gfloat * c)
{
    _trans.back().center(c);
}


/*  we need to check wheter plm != _last_motion
 *  because if another Object is running an animation, and we
 *  are pressing button 1 (so _motion == TRUE) but not
 *  moving it, Object::draw is called because of the other animation,
 *  so the current Object gets redrawn. With this aditional condition
 *  the current Object is redrawn only if we drag the mouse.
 */

GNode*
Object::moving_some_parent()
{
    GNode* ptr = _gnode;

    while(ptr) {
        if( BASE_OBJECT(ptr)->selected() &&
            BASE_OBJECT(ptr)->motion()   &&
            BASE_OBJECT(ptr)->previous_last_motion() != BASE_OBJECT(ptr)->last_motion()
          )
            return ptr;

        ptr = ptr->parent;
    }

    return NULL;
}

// The gnode returned by moving_some_parent() won't be in the
// idle_parents list, because when a object is being moving it has
// idle set to false.

gboolean
Object::some_parent_idle(GnodeList & idle_parents)
{
    GNode* ptr = _gnode;

    while(ptr) {
        if( BASE_OBJECT(ptr)->idle() )
            idle_parents.push_front(ptr);

        ptr = ptr->parent;
    }

    return idle_parents.size();
}

void
Object::apply_transformations()
{
    GnodeList idle_parents;
    Transformations*   tr;
    GNode*           node;

    for(trIter it = _trans.begin(); it != _trans.end(); it++ ) {
        glTranslatef( it->_offset[0],
                      it->_offset[1],
                      it->_offset[2] );
        rotate_about( *it );
    }

    if( node = moving_some_parent() ) {
        tr = get_transf_by_node(node);
        add_quats( & tr->lastquat(), & tr->curquat(), & tr->curquat() );

        if( node == _gnode || _gnode->next == NULL )
            BASE_OBJECT(node)->previous_last_motion( BASE_OBJECT(node)->last_motion() );
    }

    if ( some_parent_idle(idle_parents) ) {
        for(Iter it = idle_parents.begin(); it != idle_parents.end(); it++ ) {
            tr = get_transf_by_node(*it);
            add_quats( & tr->lastquat(), & tr->curquat(), & tr->curquat() );
        }
    }
}



// p is a point afected by the transformations of 'this'.
// pp will be in the "parents's coordinates"

void
Object::local_to_scene(GNode* parent, gfloat* p, gfloat* pp)
{
    gfloat m[16];
    trIter it;
    gfloat w;

    for( it = _trans.begin(); it->node() != parent; it++ );

    glPushMatrix();

    for( ; it != _trans.end(); it++ ) {
        glTranslatef( it->offset()[0],
                      it->offset()[1],
                      it->offset()[2] );
        rotate_about( *it );
    }

    glGetFloatv(GL_MODELVIEW_MATRIX, m);

    pp[0] = p[0]*m[1]  + p[1]*m[2]  + p[2]*m[3]  + m[4];
    pp[1] = p[0]*m[5]  + p[1]*m[6]  + p[2]*m[7]  + m[8];
    pp[2] = p[0]*m[9]  + p[1]*m[10] + p[2]*m[11] + m[12];
    w     = p[0]*m[13] + p[1]*m[14] + p[2]*m[14] + m[16];

    pp[0] /= w;
    pp[1] /= w;
    pp[2] /= w;

    glPopMatrix();

}


void
Object::update_bboxes()
{
    VectorArray<gfloat> tr_vertex( _vertex.size() );
    SVector<6> bbox;
    gfloat m[16];
    trIter it, it2;
    gfloat w;
    gfloat* ptr  = tr_vertex.begin();
    gfloat* ptr2 =   _vertex.begin();

    deque< SVector<6> >().swap( _bboxes );

    for( it = _trans.begin(); it != _trans.end(); it++ ) {

        // only calculate bboxes for transformations that has actuall nodes
        // asociated with it

        if( it->node() == NULL )
            continue;

        glPushMatrix();

        for( it2 = it;  it2 != _trans.end(); it2++ ) {
            glTranslatef( it2->offset()[0],
                          it2->offset()[1],
                          it2->offset()[2] );
            rotate_about( *it2 );
        }

        glGetFloatv(GL_MODELVIEW_MATRIX, m);

        for(gint i=0; i < _vertex.size()*3; i+=3) {

            ptr[i]    = ptr2[i]*m[0]  + ptr2[i+1]*m[1]  + ptr2[i+2]*m[2]  + m[3];
            ptr[i+1]  = ptr2[i]*m[4]  + ptr2[i+1]*m[5]  + ptr2[i+2]*m[6]  + m[7];
            ptr[i+2]  = ptr2[i]*m[8]  + ptr2[i+1]*m[9]  + ptr2[i+2]*m[10] + m[11];

            w         = ptr2[i]*m[12] + ptr2[i+1]*m[13] + ptr2[i+2]*m[14] + m[15];

            ptr[i]   /= w;
            ptr[i+1] /= w;
            ptr[i+2] /= w;
        }

        calc_bbox(tr_vertex,  bbox);
        _bboxes.push_back(bbox);

        glPopMatrix();
    }

    //_bbox = _bboxes.back();
    _bboxes.back() = _bbox;
}




/* Calculate the object's center */
void
Object::calc_center() {
    mean( & center(), &_vertex, _vertex.size() * 3);
}


void
Object::update_bbox()
{
    calc_bbox(_vertex,  _bbox);
}


gfloat
Object::farest_vertex(gfloat* center)
{
    gfloat radio = 0;
    util_farest_vertex( _vertex.begin(), _vertex.size(), center, &radio );
    return radio;
}


void
Object::parse_xml_node(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur)
{
    cur = cur->childs;

    while (cur != NULL) {
        if( EQSTR(cur->name, "Vertex") && (cur->ns == ns)) {
            parse_array(_vertex,
                        (gchar*) xmlNodeListGetString(doc, cur->childs, 1),
                        (gchar*) xmlGetProp(cur, "Length"));
        }
        cur = cur->next;
    }
}


void
Object::set_color_and_material_state()
{
    // TODO: this could be optimized a lot.

    if( mono() ) {
        glDisableClientState(GL_COLOR_ARRAY);
        glDisable(GL_COLOR_MATERIAL);

        if( slights.some_active() ) {
            glEnableClientState(GL_NORMAL_ARRAY);
            glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, & _material.ambiend_and_diffuse());
            glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, & _material.specular());
            glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, & _material.emission());
        }
        else {
            glDisableClientState(GL_NORMAL_ARRAY);
            glColor3fv( &_mono_color);
        }
    }
    else { //color
        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer (3, GL_FLOAT, 0, &_color);

        if( slights.some_active() ) {
            glEnable (GL_COLOR_MATERIAL);
            glEnableClientState(GL_NORMAL_ARRAY);
        }
        else
            glDisableClientState(GL_NORMAL_ARRAY);
    }
}


void
Object::draw(void)
{

    // Esta funcin lo nico que hace es prepar
    // el estado apropiado de OpenGL (aplicables a todos los objectos),
    // para que cada objecto dibuje los datos en las funciones draw
    // derivadas.
    if( show_bbox() )
        draw_bbox();

    // All objects have this array
    glVertexPointer(3, GL_FLOAT, 0, &_vertex);

    set_color_and_material_state();

    glShadeModel ( _smooth ? GL_SMOOTH : GL_FLAT );


    if(_transparent ) {
        glEnable(GL_POLYGON_STIPPLE);
        glPolygonStipple(_transparent_array);
    }
    else
        glDisable(GL_POLYGON_STIPPLE);

    // If object is moving (_idle or _motion) then use
    // '_polygon_mode_down'.

    switch( (_polygon_mode_down != POLY_NONE && (_idle || _motion)) ?
            _polygon_mode_down : _polygon_mode
          )
    {
    case POLY_FILL:
        glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
        break;

    case POLY_LINE:
        glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
        break;

    case POLY_POINT:
        glPolygonMode( GL_FRONT_AND_BACK, GL_POINT );
        break;
    default:
        break;
    }

}

GLubyte patron[][128] =
{
    {
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
        0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55
    },

    {
        0x11, 0x11, 0x11, 0x11, 0x88, 0x88, 0x88, 0x88, 0x44, 0x44, 0x44, 0x44,
        0x22, 0x22, 0x22, 0x22, 0x11, 0x11, 0x11, 0x11, 0x88, 0x88, 0x88, 0x88,
        0x44, 0x44, 0x44, 0x44, 0x22, 0x22, 0x22, 0x22, 0x11, 0x11, 0x11, 0x11,
        0x88, 0x88, 0x88, 0x88, 0x44, 0x44, 0x44, 0x44, 0x22, 0x22, 0x22, 0x22,
        0x11, 0x11, 0x11, 0x11, 0x88, 0x88, 0x88, 0x88, 0x44, 0x44, 0x44, 0x44,
        0x22, 0x22, 0x22, 0x22, 0x11, 0x11, 0x11, 0x11, 0x88, 0x88, 0x88, 0x88,
        0x44, 0x44, 0x44, 0x44, 0x22, 0x22, 0x22, 0x22, 0x11, 0x11, 0x11, 0x11,
        0x88, 0x88, 0x88, 0x88, 0x44, 0x44, 0x44, 0x44, 0x22, 0x22, 0x22, 0x22,
        0x11, 0x11, 0x11, 0x11, 0x88, 0x88, 0x88, 0x88, 0x44, 0x44, 0x44, 0x44,
        0x22, 0x22, 0x22, 0x22, 0x11, 0x11, 0x11, 0x11, 0x88, 0x88, 0x88, 0x88,
        0x44, 0x44, 0x44, 0x44, 0x22, 0x22, 0x22, 0x22
    },

    {
        0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
        0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
        0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
        0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
        0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
        0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
        0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
        0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
        0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
        0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
        0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
    },

    {
        0x00, 0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xaa, 0x00, 0x00, 0x00, 0x00,
        0xaa, 0xaa, 0xaa, 0xaa, 0x00, 0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xaa,
        0x00, 0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xaa, 0x00, 0x00, 0x00, 0x00,
        0xaa, 0xaa, 0xaa, 0xaa, 0x00, 0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xaa,
        0x00, 0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xaa, 0x00, 0x00, 0x00, 0x00,
        0xaa, 0xaa, 0xaa, 0xaa, 0x00, 0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xaa,
        0x00, 0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xaa, 0x00, 0x00, 0x00, 0x00,
        0xaa, 0xaa, 0xaa, 0xaa, 0x00, 0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xaa,
        0x00, 0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xaa, 0x00, 0x00, 0x00, 0x00,
        0xaa, 0xaa, 0xaa, 0xaa, 0x00, 0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xaa,
        0x00, 0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xaa,
    },

    {
        0x88, 0x88, 0x88, 0x88, 0x11, 0x11, 0x11, 0x11, 0x22, 0x22, 0x22, 0x22,
        0x44, 0x44, 0x44, 0x44, 0x88, 0x88, 0x88, 0x88, 0x11, 0x11, 0x11, 0x11,
        0x22, 0x22, 0x22, 0x22, 0x44, 0x44, 0x44, 0x44, 0x88, 0x88, 0x88, 0x88,
        0x11, 0x11, 0x11, 0x11, 0x22, 0x22, 0x22, 0x22, 0x44, 0x44, 0x44, 0x44,
        0x88, 0x88, 0x88, 0x88, 0x11, 0x11, 0x11, 0x11, 0x22, 0x22, 0x22, 0x22,
        0x44, 0x44, 0x44, 0x44, 0x88, 0x88, 0x88, 0x88, 0x11, 0x11, 0x11, 0x11,
        0x22, 0x22, 0x22, 0x22, 0x44, 0x44, 0x44, 0x44, 0x88, 0x88, 0x88, 0x88,
        0x11, 0x11, 0x11, 0x11, 0x22, 0x22, 0x22, 0x22, 0x44, 0x44, 0x44, 0x44,
        0x88, 0x88, 0x88, 0x88, 0x11, 0x11, 0x11, 0x11, 0x22, 0x22, 0x22, 0x22,
        0x44, 0x44, 0x44, 0x44, 0x88, 0x88, 0x88, 0x88, 0x11, 0x11, 0x11, 0x11,
        0x22, 0x22, 0x22, 0x22, 0x44, 0x44, 0x44, 0x44, }
};
