/***************************************************************************
                          scene.cc  -  description
                             -------------------
    begin                : Fri Dec 10 1999
    copyright            : (C) 1999, 2000 by Juan Pablo Romero Mndez
    email                : juan_pablo_22@yahoo.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "scene.h"
#include "surface.h"
#include "triangles.h"
#include "line.h"
#include "point.h"
#include "polygon.h"
#include "util.h"
#include "trackball.h"
#include "read_data.h"
#include "save_eps.h"


#include <GL/gl.h>
#include <GL/glu.h>

#include <setjmp.h>

jmp_buf main_loop;
thread_message t_message;

SceneLights  slights;

Scene::Scene()
{
    // this is the only node that is created by hand.
    // all other gnodes are created with Multiple::append
    _gnode = g_node_new( (Multiple*) this );
    _selected_gnode = _gnode;

    _selected = TRUE;

    _name = string("Scene");

    _idle     = FALSE;
    _animation= FALSE;
    _motion   = FALSE;
    _smooth   = FALSE;
    _normals  = FALSE;
    _mono     = TRUE;
    _normalize = FALSE;
    _axis     = TRUE;

    _two_side_lighting  = TRUE;
    _update_object_vars = TRUE;

    _bg_color   = SVector<4>(0, 0, 0, 1);

    _beginx   = 0;
    _beginy   = 0;
    _fovy     = 35;  /* degrees */
    _scale    = 1.0;
    _razon    = 1.0;
    _normals_type = 1;
    _near = 0.7;
    _far = 120;

    _x_pos  = _y_pos = _z_pos = 0.0;
    _viewpoint_inc = 1.0;
    _angle         = 90.0;

    _feedBufferSize = 100000;

    vzero( _viewpoint );

    // 
    _no_light_when_moving = FALSE;
    // 

    _timer = g_timer_new ();

    // Don't show the bbox of the hole scene
    _show_bbox = FALSE;
}


Scene::~Scene() {}


void 
Scene::remove(GNode* node)
{
    // the root object can't be deleted
    if( node == _gnode ) 
    	return;

    if( node->next )
        selected_node(node->next);
    else if ( node->prev )
        selected_node(node->prev);
    else
        selected_node(node->parent);

    Multiple::remove(node);

    if ( !some_idle() ) {
        remove_idle();
        animation(FALSE);
    }
    update_viewpoint();
}


void
Scene::selected_node(GNode* node)
{
    static GNode* prev_selected = _gnode;

    _selected_gnode = node;
    BASE_OBJECT(prev_selected)->selected(FALSE);
    BASE_OBJECT(node)->selected(TRUE);
    prev_selected = node;
}

static gboolean
_add_transformations(GNode* node, gpointer data)
{
    if( Object* ob = dynamic_cast<Object*>( BASE_OBJECT(node) ) ) {
        GNode*  ptr =  node->parent;
        // Insert the new transf. in position 0, so
        // the root object will be in pos 0 and the object's parent
        // in pos n-2 (ob is in pos n-1)
        while( ptr ) {
            ob->trans().push_front( Transformations( ptr ) );
            ptr = ptr->parent;
        }
    }
    return FALSE;
}


// add all the transformations nodes to the leaf objects
void
Scene::update_transformations(void)
{
    gint n = g_node_n_children(_gnode);
    gint n2 = g_node_n_children(_gnode->children);

    g_node_traverse( _gnode,
                     G_PRE_ORDER,
                     G_TRAVERSE_LEAFS, -1,
                     _add_transformations,
                     NULL );
}



void
Scene::update_viewpoint(void)
{
    if( is_empty() )
        return;
    
    gfloat radio = 0;
    gfloat fovyr;

    calc_center();
    radio = farest_vertex( & center() );
    
    /* make the radius 20% bigger */
    radio *= 1.2;
    /* degrees to radians */
    fovyr = _fovy * PI / 180;

    // TODO: select a better viewpoint, perhaps the same way as mathematica
    // Adems la long. del vector tiene que ser igual a _viewpoint [2].

    /* given  the field of view (fovy), as an angle in radians,
     * calculate the distance betwen the eye and the point marked as the center.
     * (ley de los senos).
     */

    _viewpoint[2] = radio * sin(PI/2 - fovyr/2) / sin(fovyr/2);
    
}



/* Read all the objects in fp.
 * If file != NULL, then fp must be NULL.
 * If fp != NULL, it is assumed that it is the
 * file descriptor of a pipe (and file should be set apropiatedly).
 */

gint
Scene::read(gchar* file, gchar* buffer, gint size)
{
    xmlDocPtr doc;
    gboolean read = FALSE;

    if( file ) {
        if( doc = parse_xml_file(file) ) {
            parse_xml_doc(doc);
            read  = TRUE;
        }
        else
            g_warning("error reading xml file: %s", file);
    }
    else {
        if( doc = xmlParseMemory(buffer, size) )
            parse_xml_doc(doc);
        else
            g_warning("error reading xml buffer");
    }

    update_transformations();
    update_viewpoint();
    update_bbox();

    return read;
}

xmlDocPtr
Scene::parse_xml_file(gchar* file)
{
    return xmlParseFile(file);
}

void
Scene::parse_xml_doc(xmlDocPtr doc)
{
    // go directly to the first group of objects
    xmlNodePtr cur = doc->root->childs;

    while (cur != NULL) {
        if( EQSTR(cur->name, "Objects") ) {
            parse_xml_object_group(doc, doc->root->ns, cur, this);
        }
        cur = cur->next;
    }
}

void
Scene::parse_xml_object_group(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur, Multiple* root)
{
    BaseObject * ob = NULL;
    gchar * linked  = xmlGetProp(cur, "Linked");

    // If current group is a multiple object
    if( EQSTR(linked, "True") ) {
        ob = new Multiple;
        ob->initialize();
        // append the group to its parent
        root->append(ob);
        // make 'ob' the current root, so the remaining
        // objects are appended to 'ob'
        root = (Multiple*) ob;
    }

    cur = cur->childs;
    while (cur != NULL) {
        if( EQSTR(cur->name, "Object") && (cur->ns == ns)) {
            ob = parse_xml_object(doc, ns, cur);
            root->append(ob);
        }
        if( EQSTR(cur->name, "Objects") ) {
            parse_xml_object_group(doc, ns, cur, root);
        }
        cur = cur->next;
    }

    // If current group is multiple
    if( EQSTR(linked, "True") ) {
        // don't show bboxes for chidrens (and itself)
        root->show_bbox_all(FALSE);
        // if current group has depth 2
        if( root->gnode()->parent == _gnode )
            // show the bbox
            root->show_bbox(TRUE);
    }

    // only select top-level objects
    if( ob->depth() == 2 )
        selected_node(ob->gnode());
    else
        selected_node(_gnode->children);

}

BaseObject*
Scene::parse_xml_object(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur)
{
    Object * ob = NULL;
    gchar * ob_type;

    // Object type
    ob_type = xmlGetProp(cur, "Type");

    if      ( EQSTR(ob_type, "QUADS"    ) ) ob = new Surface;
    else if ( EQSTR(ob_type, "LINE"     ) ) ob = new Line;
    else if ( EQSTR(ob_type, "LINES"    ) ) ob = new Line(true);
    else if ( EQSTR(ob_type, "POINT"    ) ) ob = new Point;
    else if ( EQSTR(ob_type, "TRIANGLES") ) ob = new Triangles;
    else if ( EQSTR(ob_type, "POLYGON"  ) ) ob = new Polygon;

    // go to common/specific section
    cur = cur->childs;

    while (cur != NULL) {
        // common part is parsed by Object::parse_xml_node
        if( EQSTR(cur->name, "Common") && (cur->ns == ns)) {
            ob->Object::parse_xml_node(doc, ns, cur);
        }

        if( EQSTR(cur->name, "Specific") && (cur->ns == ns)) {
            ob->parse_xml_node(doc, ns, cur);
        }
        cur = cur->next;
    }

    // initialize: common part
    ob->BaseObject::initialize();
    // object specific part
    ob->initialize();

    return ob;
}


void drawaxes()
{
    /* Next draw the Coordinate Axes */

    /* OJO: la long. de los ejes debe calcularse
     * de acuerdo al radio de la esfera que envuelve
     * a cada objeto.
     */

    static gfloat color_x[3] = {1.0, 0.0, 0.0};
    static gfloat color_y[3] = {0.0, 1.0, 0.0};
    static gfloat color_z[3] = {0.0, 0.0, 1.0};

    glLineWidth(2.0);

    glBegin(GL_LINES);

    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color_x);
    glColor3fv(color_x);
    glVertex3f(-7,0,0);
    glVertex3f(7,0,0);

    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color_y);
    glColor3fv(color_y);
    glVertex3f(0,-7,0);
    glVertex3f(0,7,0);

    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color_z);
    glColor3fv(color_z);
    glVertex3f(0, 0, -12);
    glVertex3f(0, 0, 7);

    glEnd();
}

void
Scene::draw(void)
{

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity ();

    // TODO: Cambiar esto por una traslacin adecuada.
    gluLookAt( _viewpoint[0],
               _viewpoint[1],
               _viewpoint[2],
               center()[0],
               center()[1],
               center()[2],
               0, 1, 0 );

    slights.update_position();
    /* There are some problems with the ilumination of the axes,
     * so will be deactivated until someone fixes it.
     * I think a better solution is to make the axes just another object.
     */
    if(_axis)
        //drawaxes();

        Multiple::draw();

    /* Por ahora, mover toda la escena */
    glScalef(_scale, _scale, _scale);
}

void
Scene::realize(void)
{
    // This needs to be done only once (I think).

    glClearColor(_bg_color[0],
                 _bg_color[1],
                 _bg_color[2],
                 _bg_color[3]);

    glPointSize(2);

    glEnableClientState(GL_NORMAL_ARRAY);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);


    glEnable(GL_DEPTH_TEST);

#ifdef GL_VERSION_1_2
    if( _normalize )
        glEnable( GL_RESCALE_NORMAL);
#endif
    // Esto se debe usar nicamente cuando el objeto defina sus propios
    // colores

    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glLightModeli( GL_LIGHT_MODEL_TWO_SIDE,     GL_TRUE );
    glLightModeli( GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE );

    //slights.update_diffuse();
    slights.update_all();
    slights.enable_all();


}

void
Scene::window_reshape()
{
    glViewport(0,0, window_width(), window_height());

    razon( (gfloat) window_width() / (gfloat) window_height() );

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(fovy(), razon(), near(), far());

    // Deja la matriz lista para Scene::redraw()
    glMatrixMode (GL_MODELVIEW);
}

// Returns the number that corresponds to the min z value.

// TODO: En el manual dice que minz y maxz son los valores mximos
// y mnimos de *todos* los vrtices que tienen el mismo nombre. Sin embargo
// en este caso, los valores cambian de acuerdo al punto donde se hace el click,
// aun dentro del mismo objeto. Parece que tiene que ver con la propia posicin
// del punto que "recibe" el click. As que estoy usando esto para decidir que objeto
// es el que est hasta delante, en caso de que haya varios hits. Tal vez se debe a Mesa,
// tal vez no estoy leyendo bien el manual.

static guint
processHits (GLint hits, GLuint buffer[])
{
    unsigned int i;
    GLuint names, *ptr;
    glong minz;
    gint which;
    ptr = (GLuint *) buffer;

    if( hits == 1 )
        return *(ptr + 3); //return the name

    minz = *(ptr + 1); // The first value
    which = *(ptr + 3); //The first name

    for (i = 0; i < hits; i++) { // for each hit
        names = *(ptr++);
        if( *ptr < minz ) {
            minz = *ptr;
            which  = *(ptr + 2);
        }
        ptr += 3;
    }

    return which;
}



// TODO: what is the apropiated size for this buffer ?
const gint BUFSIZE = 512;

void
Scene::select(void)
{
    GLuint selectBuf[BUFSIZE];
    GLint viewport[4];
    GLint hits;

    // clear the deque
    GnodeList().swap( Multiple::_selection_gnode_list );

    glGetIntegerv(GL_VIEWPORT, viewport);
    glSelectBuffer(BUFSIZE, selectBuf);
    glRenderMode(GL_SELECT);
    glInitNames();
    glPushName(0);

    glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();

    // The mouse y-coord goes top->bottom, but glViewport goes bottom->down,
    // thus the correct y-position is viewport[3] - event_y()

    gluPickMatrix ((GLdouble) event_x(),
                   (GLdouble) (viewport[3] - event_y()),
                   5.0, 5.0, viewport);

    // The same viewing volume
    gluPerspective(fovy(), razon(), near(), far());

    glMatrixMode(GL_MODELVIEW);
    Multiple::_selection_name_stack = 1;
    Multiple::_select_mode = TRUE;
    redraw();
    Multiple::_select_mode = FALSE;

    hits = glRenderMode (GL_RENDER);
    if( hits > 0 ) {
        gint h = processHits(hits, selectBuf);
        // The object # 1 is deque[0], so we must subtract 1 to processHits()
        //ui_notify_gnode_selected( Multiple::_selection_gnode_list[ processHits(hits, selectBuf) - 1 ] );
        ui_notify_gnode_selected( Multiple::_selection_gnode_list[ h - 1 ] );
    }

    // Restore the projection matrix
    window_reshape();
    redraw();
}


void
Scene::print(gchar * file)
{

    float a1, a2;
    GLfloat * feedBuffer;
    GLint size;
    FILE * fd;

    feedBuffer = new GLfloat[feedBufferSize()];

    glFeedbackBuffer( feedBufferSize(), GL_3D_COLOR, feedBuffer );
    glRenderMode(GL_FEEDBACK);

    redraw();

    size = glRenderMode (GL_RENDER);

    if (file) {
        fd = fopen(file, "w");
        if (fd) {
            spewWireFrameEPS(fd, 1, size, feedBuffer, "superficie 0.7.2");
        } else {
            g_warning("Could not open %s\n", file);
        }
    } else {
        printBuffer(size, feedBuffer);
    }

    delete feedBuffer;
}


void
Scene::button_press()
{
    if( is_empty() ) return;

    BaseObject* ob = selected_object();

    /* beginning of drag, reset mouse position */
    beginx( event_x() );
    beginy( event_y() );

    if( event_button() == 1 && !button2() ) {
        if( control() ) {
            select();
            return;
        }

        g_timer_start (_timer);

        ob->idle(FALSE);

        // If no object has idle(false),
        // remove the idle function
        if ( !some_idle() )
            remove_idle();

    }
}


void
Scene::motion_notify()
{
    if( is_empty() )
        return;

    BaseObject* ob = selected_object();

    if( ob == this || ob->invisible() )
        return;

    gfloat offs_[3];
    gfloat quat_[4] = {0,0,0,1};

    gint w = window_width();
    gint h = window_height();

    gint x = event_x();
    gint y = event_y();

    gfloat bx = beginx();
    gfloat by = beginy();

    if ( button1() ) {

        if ( button2() ) {
            offs_[0] = ((x - bx) / w) * 10;
            offs_[1] = ((y - by) / h) * 10;

            ob->offset( offs_);

        }
        else {
            ob->last_motion( g_timer_elapsed(_timer, NULL) );
            ob->motion(TRUE);

            trackball(quat_,
                      (2.0*bx -      w) / w,
                      (     h - 2.0*by) / h,
                      ( 2.0*x -      w) / w,
                      (     h -  2.0*y) / h);

            ob->lastquat(quat_);

            //beginx(x);
            //beginy(y);

        }
        redraw();
    }
    else if ( button2() ) {
        if ( control() ) {
            /*
            offs_[0] += ((x - bx) / w)  * 10;
            offs_[1] -= ((y - by) / h) * 10;

            ob->offset( offs_ );
            */
        }
        else {
            gfloat vp[3];

            vp[0] = viewpoint()[0];
            vp[1] = viewpoint()[1];
            vp[2] = viewpoint()[2];

            viewpoint_inc( ((y - by) / h) * 6 );

            vp[0] *= 1+viewpoint_inc();
            vp[1] *= 1+viewpoint_inc();
            vp[2] *= 1+viewpoint_inc();

            viewpoint(vp);
        }
        redraw();
    }

    beginx(x);
    beginy(y);
}



void
Scene::button_release()
{
    if( is_empty() )
        return;

    BaseObject* ob = selected_object();

    if( event_button() == 1 ) {
        if ( ob->motion() ) {
            g_timer_stop (_timer);
            if ( (g_timer_elapsed(_timer, NULL) - ob->last_motion()) < MOTION_MIN_TIME ) {
                if( !some_idle() && !is_empty() )
                    add_idle(); //una sola funcion idle (la de Scene)

                ob->idle(TRUE);
                // Significa que la posicin del objeto debe ser re-
                // calculada cada vez que draw es llamada.
            }
            ob->motion(FALSE);
        }
    }
    else if( event_button() == 2 && button1() ) {
        g_timer_start (_timer);
        ob->idle(FALSE);
    }

    update_bbox();
    redraw();
}

