
(* Title: Superficie *)

(* Context: Superficie` *)

(* Mathematica Version: 3.0 *)

(* This package provides an interface to superficie.
   There are several functions provided:
   - SupParametricPlot3D: similar to ParametricPlot3D, but doesn't display the result
           you must use SupShow[], or SupWrite[].
   - SupShow: run superficie with the given data. The data can be one of:
     a) a QUADS object
     b) a curve object
     c) Graphics3D object
     d) SurfaceGraphics object
     
     That means that to plot a Line, for example, you should say:
	
        SupShow[ Graphics3D[ Line[...] ]
     
   - SupWrite: writes a file with the given data. Accept the same input
     as SupShow
     
   - SurfaceGraphics2QUADS: converts a SurfaceGraphics object to a mesh object.
   
   The only options that this functions listen is PlotPoints. Any other option
   is discarded.
*)

(* Examples:

     To generate the data for an helicoid:
     
       helicoid = SupParametricPlot3D[ {v Cos@u, v Sin@u, u}, 
       			{u, -Pi/2, Pi/2}, 
                        {v, -5, 5}, 
                        PlotPoints->{20, 25}];
     
     To plot the data;
     
       SupShow @ helicoid;
     
     To save it to a file:
     
       SupWrite[ helicoid, "helicoid.sup"];
*)


(*
Warning: This routines doesn't check for any type of error, and
probably will change in future releases. *)

BeginPackage["Superficie`"]


(* This is the path to superficie. If it's already in your PATH, the
default is ok; alternatively you can change it to where superficie is
on your system. *)

(* for example: *)
(* $SuperficiePath = "/home/jpablo/programas/opengl/superficie/src/superficie"; *)

$SuperficiePath = "superficie";



	(* Author: Juan Pablo Romero Mndez <juan_pablo_22@yahoo.com>
           Correspondencia 146-301, Col. Postal. Del. Benito Jurez,
           Mxico D.F. C.P. 03410 Tel: 5579-5296

           *)

(* last modified: december 7, 1999 *)
(* Changes: 
    - Replace: 
    	SuperficieWrite -> SupWrite 
        SuperficieShow  -> SupShow
        mesh            -> SupParametricPlot3D
        curve           -> SupParametricPlot3D
        
      This is to homogenize names with the yacas interface
      
    - fixes the line, polygon and point stuff.
    - Change all the references of MESH to QUADS
    *)


(* Probably you don't want to change anything below here *)

(************************************************************************)

(*
    Copyright (C) 1998, 1999  by Juan Pablo Romero Mndez

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    COPYNG file for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)


SupParametricPlot3D::usage = "SupParametricPlot3D[ param, {x, xmin,
xmax}, {y, ymin, ymax} (,opts)] Generate a 3-dimensional space surface
parametrized by x (which runs from xmin to xmax) and y (from ymin to
ymax).

SupParametricPlot3D[ param, {t, tmin, tmax} (,opts)]
Generate a 3-dimensional space curve parametrized by t (which runs from
tmin to tmax).";

PlotPoints::usage = "PlotPoints is an option for superficie, and
indicates the number of points the grid for evaluating the
parametrization has. It is of the form {xpoints, ypoints}.";

SupShow::usage = "SupShow[data] Run superficie with
the specified data.";

SupWrite::usage = "SupWrite[ data, file ] Write data to
file.";

SurfaceGraphics2QUADS::usage = "SurfaceGraphics2QUADS[ SurfaceGraphics ] ";

Options[ superficie ] = { PlotPoints->15 }

Begin["`Private`"]

SupParametricPlot3D[param_, {x_, xmin_, xmax_}, {y_, ymin_, ymax_}, opts___]:=
   Module[
      {
       op = PlotPoints /. {opts} /. Options[superficie],
       datos, xinc, yinc, ptos
      },
      ptos  = If[ ListQ @ op, op, {op,op}];
      xinc  = (xmax - xmin) / (First @ ptos - 1);
      yinc  = (ymax - ymin) / (Last  @ ptos - 1);
      datos = Table[ N[param,5], {y, ymin, ymax, yinc}, {x, xmin, xmax, xinc}];

      {"QUADS", ptos, Flatten @ Transpose @ datos }
   ]

SupParametricPlot3D[param_, {t_, tmin_, tmax_}, opts___]:=
   Module[
      {
       ptos = PlotPoints /. {opts} /. Options[superficie],
       datos, xinc, yinc
      },
      tinc  = (tmax - tmin) / (ptos - 1);
      datos = Table[ N[param,5], {t, tmin, tmax, tinc}];

      {"LINE", ptos, Flatten @ datos }
   ]


writeQUADS[ datos_, id_ ]:=
   Module[{ind},
       ind = indices @@ datos[[2]];
       Write[ id, ToExpression @ "QUADS"];
       Write[ id, ToExpression @ "vertex"];
       Write[ id, Length @ Last@datos ];
       Write[ id, CForm /@ Last@datos ];
       Write[ id, ToExpression @ "index"];
       Write[ id, Length @ ind];
       Write[ id, ind ];
   ]

writePOLY[ datos_, id_ ]:=
   Module[ {fdata = Flatten @ Last @ datos},
       Write[ id, ToExpression @ "POLYGON"];
       Write[ id, ToExpression @ "vertex"];
       Write[ id, Length @ fdata ];
       Write[ id, CForm /@ fdata ];
   ]

writeLINE[ datos_, id_ ]:=
   Module[ {fdata = Flatten @ Last @ datos},
       Write[ id, ToExpression @ "LINE"];
       Write[ id, ToExpression @ "vertex"];
       Write[ id, Length @ fdata ];
       Write[ id, CForm /@ fdata ];
   ]

writePOINT[ datos_, id_ ]:=
   Module[ {fdata = Flatten @ Last @ datos},
       Write[ id, ToExpression @ "POINT"];
       Write[ id, ToExpression @ "vertex"];
       Write[ id, Length @ fdata ];
       Write[ id, CForm /@ fdata ];
   ]

writeGraphics3D[ datos_, id_ ]:=
   Module[ {fdata = Flatten @ First @ datos},
	If[ ListQ @ fdata, 
            Write[ id, ToExpression @ "MULTIPLE"];
            Write[ id, Length @ fdata];
            Switch[ Head@#,
                     Polygon, writePOLY[#, id],
                     Line,    writeLINE[#, id],
                     Point,   writePOINT[#, id]
            ]& /@ fdata,

            Switch[ Head@#,
                     Polygon, writePOLY[#, id],
                     Line,    writeLINE[#, id],
                     Point,   writePOINT[#, id]
            ]& @ fdata
        ]
   ]

SupWrite[ datos_, file_ ]:=
   Module[{id, type},
       id = OpenWrite[file];
       type = If[ ListQ @ datos,
                  First @ datos,
                  Head  @ datos
                  ];
       Switch[ type, 
              "QUADS",
              		writeQUADS[ datos, id ],
              "LINE",
              		writeLINE[ datos, id ],
              SurfaceGraphics,
              		writeQUADS[ SurfaceGraphics2QUADS @ datos, id ],
              Graphics3D,
              		writeGraphics3D[ datos, id],
              True,
              		Print["data format not supported"]
       	];
	Close[id];
   ]		


SupShow[datos_]:=
    Module[{id},
       id =  OpenTemporary[];
       Close[ id ];
       SupWrite[ datos, First@id];
       Run[$SuperficiePath, First@id, "&"];
       ]

indices[nx_, ny_]:=
   Flatten[ 
    Drop[#,-1]& /@ 
      Partition[
      Table[ 
         {i, i+1, i+nx+1, i+nx},
         {i, 0, nx  (ny - 1)}], nx ] 
    ]


SurfaceGraphics2QUADS[graphics_] :=
	Module[ {xmin, xmax, xinc, ymin, ymax, yinc, ret, ptos, datos},
		{{xmin, xmax}, {ymin,ymax}} = MeshRange /. graphics[[2]];
		ptos = 	RotateLeft @ Dimensions@ graphics[[1]];
		xinc  = (xmax - xmin)/( First@ptos -1);
		yinc  = (ymax - ymin)/(Last@ptos -1);
		ret = Table[ {x,y}, {x, xmin, xmax, xinc}, {y, ymin, ymax, yinc}];
		datos =  MapThread[
                Append[#1, #2]&, { Flatten[ret,1], Flatten @ graphics[[1]] }];
		{"QUADS", ptos, Chop[ Flatten @ datos ]}
        ]


End[]

Protect[superficie, SupShow, SupWrite, SurfaceGraphics2QUADS]

SetAttributes[ {superficie, SupShow, SupWrite, SurfaceGraphics2QUADS}, ReadProtected]

EndPackage[]
