
(* Title: Superficie *)

(* Context: Superficie` *)

(* Mathematica Version: 3.0 *)

(* This package provides an interface to superficie.
   There are several functions provided:
   - SupParametricPlot3D: similar to ParametricPlot3D, but doesn't display the result
           you must use SupShow[], or SupWrite[].
   - SupShow: run superficie with the given data. The data can be one of:
     a) a QUADS object
     b) a curve object
     c) Graphics3D object
     d) SurfaceGraphics object
     
     That means that to plot a Line, for example, you should say:
	
        SupShow[ Graphics3D[ Line[...] ]
     
   - SupWrite: writes a file with the given data. Accept the same input
     as SupShow
     
   - SurfaceGraphics2QUADS: converts a SurfaceGraphics object to a mesh object.
   
   The only options that this functions listen is PlotPoints. Any other option
   is discarded.
*)

(* Examples:

     To generate the data for an helicoid:
     
       helicoid = SupParametricPlot3D[ {v Cos@u, v Sin@u, u}, 
       			{u, -Pi/2, Pi/2}, 
                        {v, -5, 5}, 
                        PlotPoints->{20, 25}];
     
     To plot the data;
     
       SupShow @ helicoid;
     
     To save it to a file:
     
       SupWrite[ helicoid, "helicoid.sup"];
*)


(*
Warning: This routines doesn't check for any type of error, and
probably will change in future releases. *)

BeginPackage["Superficie`"]


(* This is the path to superficie. If it's already in your PATH, the
default is ok; alternatively you can change it to where superficie is
on your system. *)

(* for example: *)
(* $SuperficiePath = "/home/jpablo/programas/opengl/superficie/src/superficie"; *)

$SuperficiePath = "superficie";



	(* Author: Juan Pablo Romero Mndez <juan_pablo_22@yahoo.com>
           Correspondencia 146-301, Col. Postal. Del. Benito Jurez,
           Mxico D.F. C.P. 03410 Tel: 5579-5296

           *)

(* last modified: december 10, 1999 *)
(* Changes: 
    - Change code to generate XML
    - Replace: 
    	SuperficieWrite -> SupWrite 
        SuperficieShow  -> SupShow
        mesh            -> SupParametricPlot3D
        curve           -> SupParametricPlot3D
        
      This is to homogenize names with the yacas interface
      
    - fixes the line, polygon and point stuff.
    - Change all the references of MESH to QUADS
    *)


(* Probably you don't want to change anything below here *)

(************************************************************************)

(*
    Copyright (C) 1998, 1999  by Juan Pablo Romero Mndez

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    COPYNG file for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)


SupParametricPlot3D::usage = "SupParametricPlot3D[ param, {x, xmin,
xmax}, {y, ymin, ymax} (,opts)] Generate a 3-dimensional space surface
parametrized by x (which runs from xmin to xmax) and y (from ymin to
ymax).

SupParametricPlot3D[ param, {t, tmin, tmax} (,opts)] Generate a
3-dimensional space curve parametrized by t (which runs from tmin to
tmax).";

PlotPoints::usage = "PlotPoints is an option for superficie, and
indicates the number of points the grid for evaluating the
parametrization has. It is of the form {xpoints, ypoints}.";

SupShow::usage = "SupShow[data] Run superficie with the specified
data.";

SupWrite::usage = "SupWrite[ data, file ] Write data to file.";

SurfaceGraphics2QUADS::usage = "SurfaceGraphics2QUADS[ SurfaceGraphics] ";

Options[ superficie ] = { PlotPoints->15 }

Begin["`Private`"]

SupParametricPlot3D[param_, {x_, xmin_, xmax_}, {y_, ymin_, ymax_}, opts___]:=
   Module[
      {
       op = PlotPoints /. {opts} /. Options[superficie],
       datos, xinc, yinc, ptos
      },
      ptos  = If[ ListQ @ op, op, {op,op}];
      xinc  = (xmax - xmin) / (First @ ptos - 1);
      yinc  = (ymax - ymin) / (Last  @ ptos - 1);
      datos = Table[ N[param,5], {x, xmin, xmax, xinc}, {y, ymin, ymax, yinc}];

      {"QUADS", ptos, Chop @ Flatten @ Transpose @ datos }
   ]

SupParametricPlot3D[param_, {t_, tmin_, tmax_}, opts___]:=
   Module[
      {
       ptos = PlotPoints /. {opts} /. Options[superficie],
       datos, xinc, yinc
      },
      tinc  = (tmax - tmin) / (ptos - 1);
      datos = Table[ N[param,5], {t, tmin, tmax, tinc}];

      {"LINE", ptos, Chop @ Flatten @ datos }
   ]






supListToString[lst_List]:= 
   StringReplace[ 
   	StringTake[#, {2, StringLength @ # - 1}] & @ ToString @ lst,
	","->""
        ]

writeQUADS[ datos_, id_ ]:=
   Module[{ind, n1, n2, n3},
       ind = indices @@ datos[[2]];

       n1 = xmlNewNode["sup:Vertex"];
       xmlSetProp[n1, "Length", ToString@Length@Last@datos];
       xmlSetContent[n1, supListToString[CForm /@ Last@datos]];

       n2 = xmlNewNode["sup:Common"];
       xmlInsertChild[n2, n1];
       
       n1 = xmlNewNode["sup:Index"];
       xmlSetProp[n1, "Length", ToString@Length@ind];
       xmlSetContent[n1, supListToString@ind];
       
       n3 = xmlNewNode["sup:Specific"];
       xmlInsertChild[n3, n1];
       
       n1 = xmlNewNode["sup:Object"];
       xmlSetProp[n1, "Type", "QUADS"];
       xmlInsertChild[n1, n2];
       xmlInsertChild[n1, n3];
       
       n2 = xmlNewNode["sup:Objects"];
       xmlInsertChild[n2, n1];
       
       n1 = xmlNewNode["sup:Scene"];
       xmlSetProp[n1, "xmlns:sup", "http://www.linuxsupportline.com/~superficie/"];
       xmlInsertChild[n1, n2];
      
       xmlDocDumpFile[id, n1];
   ]


writePOLY[ datos_, id_ ]:=
   Module[ {fdata = Flatten @ Last @ datos, n1, n2},

       n1 = xmlNewNode["sup:Vertex"];
       xmlSetProp[n1, "Length", ToString@Length@fdata];
       xmlSetContent[n1, supListToString[CForm /@ fdata]];
       
       n2 = xmlNewNode["sup:Common"];
       xmlInsertChild[n2, n1];
       
       n1 = xmlNewNode["sup:Object"];
       xmlSetProp[n1, "Type", "POLYGON"];
       xmlInsertChild[n1, n2];
      
       
       n2 = xmlNewNode["sup:Objects"];
       xmlInsertChild[n2, n1];
       
       n1 = xmlNewNode["sup:Scene"];
       xmlSetProp[n1, "xmlns:sup", "http://www.linuxsupportline.com/~superficie/"];
       xmlInsertChild[n1, n2];
      
       xmlDocDumpFile[file, n1];

   ]

writeLINE[ datos_, id_ ]:=
   Module[ {fdata = Flatten @ Last @ datos, n1, n2},

       n1 = xmlNewNode["sup:Vertex"];
       xmlSetProp[n1, "Length", ToString@Length@fdata];
       xmlSetContent[n1, supListToString[CForm /@ fdata]];
       
       n2 = xmlNewNode["sup:Common"];
       xmlInsertChild[n2, n1];
       
       n1 = xmlNewNode["sup:Object"];
       xmlSetProp[n1, "Type", "LINE"];
       xmlInsertChild[n1, n2];

       n2 = xmlNewNode["sup:Objects"];
       xmlInsertChild[n2, n1];
       
       n1 = xmlNewNode["sup:Scene"];
       xmlSetProp[n1, "xmlns:sup", "http://www.linuxsupportline.com/~superficie/"];
       xmlInsertChild[n1, n2];
      
       xmlDocDumpFile[file, n1];
       ]





writePOINT[ datos_, id_ ]:=
   Module[ {fdata = Flatten @ Last @ datos, n1, n2},

       n1 = xmlNewNode["sup:Vertex"];
       xmlSetProp[n1, "Length", ToString@Length@fdata];
       xmlSetContent[n1, supListToString[CForm /@ fdata]];
       
       n2 = xmlNewNode["sup:Common"];
       xmlInsertChild[n2, n1];
       
       n1 = xmlNewNode["sup:Object"];
       xmlSetProp[n1, "Type", "POINT"];
       xmlInsertChild[n1, n2];

       n2 = xmlNewNode["sup:Objects"];
       xmlInsertChild[n2, n1];
       
       n1 = xmlNewNode["sup:Scene"];
       xmlSetProp[n1, "xmlns:sup", "http://www.linuxsupportline.com/~superficie/"];
       xmlInsertChild[n1, n2];
      
       xmlDocDumpFile[file, n1];
       ]


(* Needs to be rewritten *)
(*
writeGraphics3D[ datos_, id_ ]:=
   Module[ {fdata = Flatten @ First @ datos},
	If[ ListQ @ fdata, 
            Write[ id, ToExpression @ "MULTIPLE"];
            Write[ id, Length @ fdata];
            Switch[ Head@#,
                     Polygon, writePOLY[#, id],
                     Line,    writeLINE[#, id],
                     Point,   writePOINT[#, id]
            ]& /@ fdata,

            Switch[ Head@#,
                     Polygon, writePOLY[#, id],
                     Line,    writeLINE[#, id],
                     Point,   writePOINT[#, id]
            ]& @ fdata
        ]
   ]
*)



SupWrite[ datos_, file_ ]:=
   Module[{id, type},
       id = OpenWrite[file];
       type = If[ ListQ @ datos,
                  First @ datos,
                  Head  @ datos
                  ];
       Switch[ type, 
              "QUADS",
              		writeQUADS[ datos, id ],
              "LINE",
              		writeLINE[ datos, id ],
              SurfaceGraphics,
              		writeQUADS[ SurfaceGraphics2QUADS @ datos, id ],
              True,
              		Print["data format not supported"]
       	];
	Close[id];
   ]		


SupShow[datos_]:=
    Module[{id},
       id =  OpenTemporary[];
       Close[ id ];
       SupWrite[ datos, First@id];
       Run[$SuperficiePath, First@id, "&"];
       ]

indices[nx_, ny_]:=
   Flatten[ 
    Drop[#,-1]& /@ 
      Partition[
      Table[ 
         {i, i+1, i+nx+1, i+nx},
         {i, 0, nx  (ny - 1)}], nx ] 
    ]


SurfaceGraphics2QUADS[graphics_] :=
	Module[ {xmin, xmax, xinc, ymin, ymax, yinc, ret, ptos, datos},
		{{xmin, xmax}, {ymin,ymax}} = MeshRange /. graphics[[2]];
		ptos = 	RotateLeft @ Dimensions@ graphics[[1]];
		xinc  = (xmax - xmin)/( First@ptos -1);
		yinc  = (ymax - ymin)/(Last@ptos -1);
		ret = Table[ {x,y}, {x, xmin, xmax, xinc}, {y, ymin, ymax, yinc}];
		datos =  MapThread[
                Append[#1, #2]&, { Flatten[ret,1], Flatten @ graphics[[1]] }];
		{"QUADS", ptos, Chop[ Flatten @ datos ]}
        ]





(* XML helper functions *)


xmlNewNode[name_String]:=
Module[ {lst, props},
   lst["name"] = name;
   lst["props"] = props;
   lst["propsLst"] = {};
   lst["content"] = "";
   lst["childs"] = {};
   lst
]

xmlSetName[node_Symbol, name_String]:= node["name"] := name;
xmlGetName[node_Symbol] := node["name"];

xmlSetProp[node_Symbol, prop_String, val_String]:=
Module[{},
 node["props"][prop] = val;
 node["propsLst"] = Insert[ node["propsLst"], prop, Length[node["propsLst"]]+1 ];
]


xmlGetProp[node_Symbol, prop_String] := node["props"][prop];
xmlSetContent[node_Symbol, data_String]:= node["content"] = data;

xmlGetContent[node_Symbol]:=  node["content"];

xmlInsertChild[node_Symbol, child_Symbol] :=
 node["childs"] = Insert[ node["childs"], child, Length[node["childs"]]+1 ];

xmlGetChild[node_Symbol, n_Integer] := node["childs"][[n]];

xmlDocDumpFile[file_, node_Symbol] :=
Module [{ helper, id}, 
   helper[id_, node_] :=
      Module[ {tmp},
       WriteString[id, "<"];
       WriteString[id, xmlGetName@node];
       tmp = ( (" " <> # <> "=\"" <> node["props"][#]  <> "\"")& /@  node["propsLst"]);
       Scan[ WriteString[id, #]&, tmp];
       WriteString[id,">\n"];
       Scan[ helper[id,#]&, node["childs"] ];
       WriteString[id, xmlGetContent@node];
       If[ xmlGetContent@node != "", WriteString[id,"\n"]];
       WriteString[id, "</" <> xmlGetName @ node <> ">\n"];
      ];
        
   If[ StringQ@file,
       id = OpenWrite[file],
       id = file
       ];

   WriteString[ id, "<?xml version=\"1.0\"?>\n"];
   helper[id, node];

   If[ StringQ@file,
       Close[id]
       ];
]





End[]

Protect[SupParametricPlot3D, SupShow, SupWrite, SurfaceGraphics2QUADS]

SetAttributes[ {SupParametricPlot3D, SupShow, SupWrite, SurfaceGraphics2QUADS}, ReadProtected]

EndPackage[]
