## usage:  surf_sf (XX, YY, ZZ) 
##         surf_sf (XX, YY) 
##         surf_sf (ZZ) 
##         where XX and YY can be vectors and matrices and ZZ must be a
##         matrix
##         XX and YY contain the x and y values for the mesh to be
##         plotted, ZZ containes the elevation
##         if XX is a vector, then length(XX) must be equal to columns(ZZ)
##         if ZZ is a vector, then length(ZZ) must be equal to rows(ZZ)
##         if XX or YY is a matrix, then it must have the same size as ZZ

## Copyright (C) 1999 Peter Hopfgartner
##
## This file is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2, or (at your option)
## any later version.
##
## This file is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with Octave; see the file COPYING.  If not, write to the Free
## Software Foundation, 59 Temple Place - Suite 330, Boston, MA
## 02111-1307, USA.

## Author: Peter Hopfgartner <hopfgartner@rolmail.net>
##
## Changes: December 1999 Changed to XML output and greatly improved
##          input check and flexibility, added small description
## Created: November 1999

function surf_sf (XX, YY, ZZ)

  if (nargin == 1)

    ## Plot on uniform grid
    ## assume that the only argument is the elevation
    if (is_matrix(XX))

      ZZ = XX;
      ## Construct the XX and YY matrices with unitary step
      XX = ones(1,rows(ZZ))'*[1:columns(ZZ)];
      YY = [1:rows(ZZ)]'*ones(1,columns(ZZ));
    else
      error ("argument must be a matrix");
    endif

  elseif (nargin == 2 || nargin == 3)

    if (is_matrix(XX) && is_matrix(YY))
      if (is_vector(YY))
	## Transform YY is a matrix containing all columns = YY
	if (is_vector(XX))
	  YY = YY([1:length(YY)])'*ones(1, length (XX));
	else
	  YY = YY([1:length(YY)])'*ones(1, columns (XX));
	endif
      endif
      if (is_vector(XX))
	## Transform XX is a matrix containing all rows = XX
	XX = ones(rows(YY),1)*XX([1:length(XX)]);
      endif
    else
      error ("arguments must be a matrices");
    endif

    if (size(XX) != size(YY))
      error ("matrices must have the same size");
    endif

    if (nargin == 2)
      ## if nargin == 2, plot mesh only and assume that
      ## the two arguments are the coordinates of the mesh
      ZZ = zeros(size(XX));
    else
      if (is_matrix(ZZ))
	if (size(XX) != size(ZZ))
	  error ("mismatch between the matrix dimensions");
	endif
      else
	error ("third argument must be a  matrix");
      endif
    endif
  else
    error ("1, 2 or 3 arguments required");
  endif
    
  [xrow xcol] = size (XX);
  tmp_in_file = tmpnam();
  [fid, msg] = fopen(tmp_in_file, "w");
  if (fid == -1)
    error("while opening superficie input file: \n\t%s", msg);
  endif

  ## Write XML preamble
  fprintf (fid, "<?xml version=\"1.0\"?>\n");
  fprintf (fid, "<sup:Scene xmlns:sup=");
  fprintf (fid, "\"http://www.linuxsupportline.com/~superficie/\">\n");
  fprintf (fid, "  <sup:Objects>\n");
  fprintf (fid, "    <sup:Object Type=\"QUADS\">\n");
  fprintf (fid, "      <sup:Common>\n");

  ## Write Vertices
  fprintf (fid, "        <sup:Vertex Length=\"%d\">\n",\
	   3 * xrow * xcol);
  for i = 1:xrow
    for j = 1:xcol
      fprintf (fid, "          %f  %f  %f\n", XX(i, j), YY(i, j), ZZ(i, j));
    endfor
  endfor
  fprintf (fid, "        </sup:Vertex>\n");
  fprintf (fid, "      </sup:Common>\n");

  ## Write Connectivity Indeces
  fprintf (fid, "      <sup:Specific>\n");
  fprintf (fid, "        <sup:Index Length=\"%d\">\n", \
	   4 * (xrow-1) * (xcol-1));
  for i=1:xrow-1
    for j=1:xcol-1
      i1 = (i - 1) * xcol + j - 1;
      i2 = (i - 1) * xcol + j;
      i3 = i * xcol + j;
      i4 = i * xcol + j - 1;
      fprintf ( fid, "          %d  %d  %d  %d\n", i1, i2, i3, i4);
    endfor
  endfor
  fprintf (fid, "        </sup:Index>\n");
  fprintf (fid, "      </sup:Specific>\n");
  fprintf (fid, "    </sup:Object>\n");
  fprintf (fid, "  </sup:Objects>\n");
  fprintf (fid, "</sup:Scene>\n");
  fflush (fid);
  fclose (fid);
  command_string = sprintf('superficie  %s > superficie.log',tmp_in_file);
  system (command_string, 'sync');
  command_string = sprintf("rm %s", tmp_in_file);
  system(command_string);
endfunction